function[splits, assignments, cutoffs, tmi] = quantize(expr)
%% Arguments: 
%% 'expr': an MxN matrix, which represents the expression (real-values) of
%% M genes in N experiments.
%%
%% Returned:
%% 'splits': an integer, denoting the "optimal" number of levels we found
%% 'assignments': an MxN matrix, paralleling the structure of 'expr',
%% except that instead of expression values, (i,j) indicates *which* level
%% (by index) gene i is in, in expression experiment j.  (i.e., the value
%% of (i,j) should be in [1, splits]).
%% 'cutoffs': a matrix of dimension, (splits-1)x N.  Each column
%% corresponds to a different experiment.  (i,j) contains an estimated
%% cutoff (a real-number) between "level i" and "level i+1" in experiment
%% j.  (that is, cutoffs(i,j) = 0.5 * (max(expr(level i genes of expt j)) +
%% min(expr(level i+1 genes of expt j))) )
%% 'tmi' and 'diff_tmi' are two arrays of values.  'tmi' can be used to
%% generate the plots of TMI vs. Level that were found in the notes.
%% 'diff_tmi' is the array of pairwise differences, which is minimized.

expts = size(expr, 2);

%% 'total_A' is a 3-D matrix which describes the quantization matrices
%% for each different variable.  It has dimensions LxMxN if there 
%% are L different variables (expression experiments), M different 
%% genes, and N different "quantization levels."  The element (i,j,k)
%% == 1 if gene j's expression in experiment i is in the k^th 
%% quantization division.
total_A = build_total_A(expr);

base_pmi = pairwise_mutual_info(total_A);
levels = size(total_A{expts}, 2);
extreme_diff = 0.0;
extreme_total_A = total_A;
tmi(levels-2, :) = [levels, base_pmi, 0.0];
first = 1;

%% We have to iterate through *all* the possible reductions; 
%% At the end, we take the best one (arg min_i {TMI(i) - TMI(i-1)})
while levels > 2
    
    %% (the call to mutual_info is within this function)
    [pmi, new_total_A] = reduce_step(base_pmi, total_A);
    
    %diff = pmi - base_pmi;
    diff = base_pmi - pmi;
    
    if first == 1 || diff > extreme_diff
        extreme_total_A = total_A;
        extreme_diff = diff;
    end
        
    tmi(levels-2, :) = [levels, pmi, diff];
    
    total_A = new_total_A;
    base_pmi = pmi;
    levels = size(total_A{expts}, 2);
    first = 0;
end

%% when we're finished, find the minimum diff_tmi and call the appropriate
%% helper functions to assemble the return values.
splits = size(extreme_total_A{expts}, 2);
fprintf(1, '\nOptimal Splits: %d (diff-TMI: %f)\n', splits, extreme_diff);

assignments = get_assignments(extreme_total_A);
cutoffs = read_all_cutoffs(extreme_total_A, expr);